/* Vector gfx test
 *
 * by Niels Mller 1995
 *
 * This test program uses the functions from vectors.c to display a cube.
 * Coordinates are transformed, and a simple "backside cull" is done before
 * drawing it on screen */

#include <vectors.h>
#include <stdio.h>
#include <math.h>

typedef int Triangle[3];
 
/* Data */
#define NUMBER_OF_POINTS 8
#define NUMBER_OF_TRIANGLES 12

Vector my_points[NUMBER_OF_POINTS] =
{
   {0,0,0},{0,0,1},{0,1,1},{0,1,0},
   {1,1,0},{1,0,0},{1,0,1},{1,1,1}
};

Vector light = { 10,50,20 };   /* Light source */

Triangle my_triangles[NUMBER_OF_TRIANGLES] = 
{
   { 0,1,2 },
   { 0,2,3 },
   { 0,3,4 },
   { 0,4,5 },   
   { 0,5,6 },
   { 0,6,1 },
   { 7,2,1 },
   { 7,3,2 },
   { 7,4,3 },
   { 7,5,4 },
   { 7,6,5 },
   { 7,1,6 }
};

struct Viewpoint view =
{
  {0.5,1,1},       /* Reference point */
  {-20,-40,-100},  /* Direction */
  {0,50,0},        /* Upward */
  3,               /* Distance */
  {0,0,2}          /* Centre */
};

/* Area of projection plane to display */
#define MINX -.7
#define MAXX .7
#define MINY -0.5
#define MAXY 0.5
 
void render(int tn, Triangle triangles[], int pn, Vector points[],
	    struct Viewpoint *view, Vector l);

/* These three functions are system dependent.
 * Amiga versions are included at the end of this file. */

void openAll(void);
void closeAll(void);
void drawTriangle(Triangle t, Vector c[], double brightness); /* 0 <= brightness <= 1 */

int main(int argc, char **argv)
{
  int c;
  
  if (atexit(&closeAll))
    exit(20);
  openAll();
  render(NUMBER_OF_TRIANGLES, my_triangles, NUMBER_OF_POINTS, my_points, &view, light);
  printf("Press return.\n");
  scanf("%c",&c);
  return 0;
}


void render(int tn, Triangle triangles[], int pn, Vector points[],
	    struct Viewpoint *view, Vector light)
{
  static Matrix T;
  static Vector transformed_points[NUMBER_OF_POINTS]; /* Ugly, but variably sized arrays are not standard */
  int i;
  Vector l;
  
  /* Compute vector from reference point to light source, and normalize. */
  {
    int i;
    double d;
    
    for (i=0; i<3; i++)
      l[i] = light[i] - view->ref[i];

    d = sqrt( SQR(l[0]) + SQR(l[1]) + SQR(l[2]) );
    l[0] /= d; l[1] /= d; l[2] /= d;
  }

  MakeTransform(view, T);
  {  /* Print matrix */
    int i,j;
    printf("Transformation matrix:\n");
    for (i=0; i<4; i++)
      {
	for (j=0; j<4; j++)
	  printf("  %lf", T[i][j]);
	printf("\n");
      }
  }
  TransformVectors(T, pn, points, transformed_points);

  printf("Transformed points:\n");
  for (i=0; i<pn; i++)
    printf("( %lf, %lf, %lf )\n", transformed_points[i][0],
	   transformed_points[i][1], transformed_points[i][2]);
  
  for (i = 0; i<tn; i++)
    {
      COORD_T x1,x2,y1,y2,z1,z2;

      printf("Processing triangle %d\n",i);
      /* Don't render invisible triangles. Compute two edge vectors,
       * and their cross product. */
      x1 = transformed_points [triangles[i][1]] [0]
	- transformed_points [triangles[i][0]] [0];
      y1 = transformed_points [triangles[i][1]] [1]
	- transformed_points [triangles[i][0]] [1];
      x2 = transformed_points [triangles[i][2]] [0]
	- transformed_points [triangles[i][1]] [0];
      y2 = transformed_points [triangles[i][2]] [1]
	- transformed_points [triangles[i][1]] [1];

      /* Test z-component of cross product; if it is positive then the
       * surface normal points towards the viewer. */
      if (0 < (x1*y2 - x2*y1))
	{
	  COORD_T nx,ny,nz;
	  int j;
	  double c;
	  
	  /* Compute scalar product of normal and light direction,
	   * to determine brightness. This is done in the original "world" coordinates.
	   */
	  printf("Visible.\n");
	  x1 = points [triangles[i][1]] [0] - points [triangles[i][0]] [0];
	  y1 = points [triangles[i][1]] [1] - points [triangles[i][0]] [1];
	  z1 = points [triangles[i][1]] [2] - points [triangles[i][0]] [2];
	  x2 = points [triangles[i][2]] [0] - points [triangles[i][1]] [0];
	  y2 = points [triangles[i][2]] [1] - points [triangles[i][1]] [1];
	  z2 = points [triangles[i][2]] [2] - points [triangles[i][1]] [2];

	  /* Cross product gives the normal. */
	  nx = y1*z2 - y2*z1;
	  ny = z1*x2 - z2*x1;
	  nz = x1*y2 - x2*y1;

	  /* Scalar product gives cos(alfa), where alfa is the direction between
	   * the ligth rays and the surface normal. */
	  c = (nx*l[0] + ny*l[1] + nz*l[2])
	    / sqrt( (SQR(nx) + SQR(ny) + SQR(nz)) );
	  
	  /* There is light on the surface if cos (alfa) > 0. */
	  if (c<0)
	    c = 0;
	  drawTriangle(triangles[i], transformed_points, c);
	}
    }
}

/* System dependent code and variables follow */

#ifdef AMIGA

#include <intuition/intuition.h>
#include <intuition/screens.h>
#include <graphics/rastport.h>
#include <proto/graphics.h>
#include <proto/intuition.h>

/* Size in words, gives 20 = 4*5 bytes */
#define AREA_SIZE 10

struct Screen *screen = NULL;
struct Window *window = NULL;
struct RastPort *rp;
PLANEPTR planePtr = NULL;

struct ColorSpec colors[] =
{
  {0,0,0,0},
  {1,15,10,12}, 
  {2,2,2,2},
  {3,3,3,3},
  {4,4,4,4},
  {5,5,5,5},
  {6,6,6,6},
  {7,7,7,7},
  {8,8,8,8},
  {9,9,9,9},
  {10,10,10,10},
  {11,11,11,11},
  {12,12,12,12},
  {13,13,13,13},
  {14,14,14,14},
  {15,15,15,15},
  {-1,0,0,0}
};

#define SCALEX(x) \
  ( (SHORT) window->Width * ( ((x) - MINX) / (MAXX - MINX)) )

#define SCALEY(y) \
  ( (SHORT) window->Height * ( (MAXY - (y)) / (MAXY - MINY)) ) 

void drawTriangle(Triangle t, Vector c[], double brightness)
{
  SetAPen(rp, (int) (2 + 13.9 * brightness));
  AreaMove(rp, SCALEX( c[t[0]][0]), SCALEY( c[t[0]][1]));
  AreaDraw(rp, SCALEX( c[t[1]][0]), SCALEY( c[t[1]][1]));
  AreaDraw(rp, SCALEX( c[t[2]][0]), SCALEY( c[t[2]][1]));
  AreaEnd(rp);
}

void openAll(void)
{
   static WORD areaBuffer[AREA_SIZE];
   static struct AreaInfo areainfo;
   static struct TmpRas tmpRas;

  if (!(screen = OpenScreenTags(NULL, SA_Depth, 4,
				SA_Colors, (LONG) colors,
                                SA_Type, (LONG) CUSTOMSCREEN,
				SA_Title, (LONG) "Vectors",
                                SA_SysFont, 1,
                                TAG_DONE))) 
    {
      printf("Could not open screen.");
      exit(10);
    }
  if (!(window = OpenWindowTags(NULL, WA_ScreenTitle, (LONG) "Vector gfx test",
                                WA_CustomScreen, (LONG) screen,
                                WA_Top, screen->BarHeight+2,
                                WA_Left, 0,
                                WA_Width, screen->Width,
                                WA_Height, screen->Height - (screen->BarHeight+2),
                                WA_Flags, WFLG_SMART_REFRESH
                                | WFLG_BORDERLESS | WFLG_NOCAREREFRESH,
                                TAG_DONE )))
    {
      printf("Could not open window.\n");
      exit(10);
    }

   rp = window->RPort;

   if (!(planePtr = AllocRaster(window->Width, window->Height) ))
     {
       printf("Could not allocate raster.\n");
       exit(10);
     }
   InitArea(&areainfo, areaBuffer, (AREA_SIZE*2)/5 );
   rp->AreaInfo = &areainfo;
   InitTmpRas(&tmpRas, planePtr, RASSIZE(window->Width, window->Height) );
   rp->TmpRas = &tmpRas;
}

void closeAll(void)
{
  if (planePtr)
    FreeRaster(planePtr, window->Width, window->Height);
  if (window)
    CloseWindow(window);
  if (screen)
    CloseScreen(screen);
}

#endif AMIGA
